# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import LINUX_LIKE

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class EOPKG(PackageManager):
    name = "Solus package manager"

    homepage_url = "https://github.com/getsolus/eopkg/"

    platforms = LINUX_LIKE

    requirement = "3.2.0"

    pre_args = ("--no-color",)

    _LIST_REGEXP = re.compile(
        r"^(?P<package_id>\S+)\s+\|\.+\|\s+(?P<version>\.+)\|\.+\|\.+\|\.+$",
    )
    _SEARCH_REGEXP = re.compile(
        r"^(?P<package_id>\S+)\s+- (?P<description>\.+)$",
    )

    version_regexes = (r"eopkg\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ eopkg --version
        eopkg 3.2.0
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ eopkg --no-color list-installed --install-info
            Package Name          |St|        Version|  Rel.|  Distro|       Date
            =====================================================================
            aalib                 | i|        1.4.0_5|     8|   Solus|14 Oct 2024
            abseil-cpp            | i|     20240116.2|    10|   Solus|14 Oct 2024
            accountsservice       | i|        23.13.9|    36|   Solus|14 Oct 2024
            acl                   | i|          2.3.2|    21|   Solus|14 Oct 2024
            adwaita-icon-theme    | i|           46.2|    28|   Solus|14 Oct 2024
            adwaita-icon-theme-legacy  | i|           46.2|     2|   Solus|14 Oct 2024
            alsa-firmware         | i|          1.2.4|     7|   Solus|14 Oct 2024
            alsa-lib              | i|         1.2.12|    38|   Solus|14 Oct 2024
            alsa-plugins          | i|         1.2.12|    26|   Solus|14 Oct 2024
            alsa-utils            | i|         1.2.12|    28|   Solus|14 Oct 2024
            aom                   | i|         3.10.0|    24|   Solus|14 Oct 2024
            appstream             | i|          1.0.1|     9|   Solus|14 Oct 2024
            appstream-data        | i|             49|    51|   Solus|14 Oct 2024
            appstream-glib        | i|          0.8.2|    13|   Solus|14 Oct 2024
            argon2                | i|       20190702|     6|   Solus|14 Oct 2024
            at-spi2               | i|         2.52.0|    44|   Solus|14 Oct 2024
            atkmm                 | i|         2.28.4|    19|   Solus|14 Oct 2024
            attr                  | i|          2.5.2|    25|   Solus|14 Oct 2024
            audit                 | i|          4.0.2|    19|   Solus|14 Oct 2024
            avahi                 | i|            0.8|    27|   Solus|14 Oct 2024
            baobab                | i|           46.0|    27|   Solus|14 Oct 2024
        """
        output = self.run_cli("list-installed", "--install-info")

        for package in output.splitlines()[:-2]:
            match = self._LIST_REGEXP.match(package)
            if match:
                package_id, installed_version = match.groups()
                yield self.package(id=package_id, installed_version=installed_version)

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ eopkg --no-color list-upgrades --install-info
            Package Name          |St|        Version|  Rel.|  Distro|       Date
            =====================================================================
            adwaita-icon-theme   | i|           46.2|    28|   Solus|14 Oct 2024
            appstream-data       | i|             49|    51|   Solus|14 Oct 2024
            at-spi2              | i|         2.52.0|    44|   Solus|14 Oct 2024
            baobab               | i|           46.0|    27|   Solus|14 Oct 2024
        """
        output = self.run_cli("list-upgrades", "--install-info")

        for package in output.splitlines()[:-2]:
            match = self._LIST_REGEXP.match(package)
            if match:
                package_id, installed_version = match.groups()
                yield self.package(id=package_id, installed_version=installed_version)

    @search_capabilities(exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not supports exact matching.

        Naked search without parameters is the same as extended search with all filtering
        parameters (i.e. ``--name --summary --description``):

        .. code-block:: shell-session

            $ eopkg --no-color search firefox
            gjs-dbginfo                 - Debug symbols for gjs
            bleachbit                   - BleachBit frees disk space and maintains privacy
            firefox                     - Firefox web browser
            eid-mw-firefox              - Belgian eID add-on for Mozilla Firefox
            gjs                         - GNOME JavaScript
            font-fira-ttf               - Mozilla's new typeface, used in Firefox OS
            geckodriver                 - WebDriver for Firefox
            firefox-dbginfo             - Debug symbols for firefox
            nvidia-vaapi-driver-dbginfo - Debug symbols for nvidia-vaapi-driver
            font-clear-sans-ttf         - Clear Sans Fonts - TrueType
            gjs-devel                   - Development files for gjs
            geckodriver-dbginfo         - Debug symbols for geckodriver

            $ eopkg --no-color search firefox --name --summary --description
            gjs-dbginfo                 - Debug symbols for gjs
            bleachbit                   - BleachBit frees disk space and maintains privacy
            firefox                     - Firefox web browser
            eid-mw-firefox              - Belgian eID add-on for Mozilla Firefox
            gjs                         - GNOME JavaScript
            font-fira-ttf               - Mozilla's new typeface, used in Firefox OS
            geckodriver                 - WebDriver for Firefox
            firefox-dbginfo             - Debug symbols for firefox
            nvidia-vaapi-driver-dbginfo - Debug symbols for nvidia-vaapi-driver
            font-clear-sans-ttf         - Clear Sans Fonts - TrueType
            gjs-devel                   - Development files for gjs
            geckodriver-dbginfo         - Debug symbols for geckodriver

        For default search on package name only, we rescript filtering to ``--name`` only:

        .. code-block:: shell-session

            $ eopkg --no-color search firefox --name
            firefox         - Firefox web browser
            eid-mw-firefox  - Belgian eID add-on for Mozilla Firefox
            firefox-dbginfo - Debug symbols for firefox
        """
        # Extended search is the default behavior.
        arg = ""
        # Non-extended search restrict matching to package name only.
        if not extended:
            arg = "--name"

        output = self.run_cli("search", arg, query)

        for package_id, description in self._SEARCH_REGEXP.findall(output):
            yield self.package(id=package_id, description=description)

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ sudo eopkg --no-color install --yes-all 0ad
            Warning: Updates available, checking reverse dependencies of runtime dependencies for safety.
            Following packages will be installed:
            0ad          0ad-data         assimp           at-spi2             baobab             breeze-icons      budgie-control-center  budgie-desktop      dav1d                  enet                   evolution-data-server       ffmpeg
            file-roller  firefox          fmt              fontconfig          gcr-4              gloox             gnome-calculator       gnome-calendar      gnome-online-accounts  gnome-settings-daemon  gnome-system-monitor        gnome-terminal
            gvfs         harfbuzz         ibus             kf6-karchive        kf6-kauth          kf6-kbookmarks    kf6-kcodecs            kf6-kcolorscheme    kf6-kcompletion        kf6-kconfig            kf6-kconfigwidgets          kf6-kcoreaddons
            kf6-kcrash   kf6-kdbusaddons  kf6-kded         kf6-kdoctools       kf6-kglobalaccel   kf6-kguiaddons    kf6-ki18n              kf6-kiconthemes     kf6-kio                kf6-kitemviews         kf6-kjobwidgets             kf6-knotifications
            kf6-kparts   kf6-kservice     kf6-kwallet      kf6-kwidgetsaddons  kf6-kwindowsystem  kf6-kxmlgui       kf6-solid              kpmcore             ldb                    libadwaita             libarchive                  libass
            libcheese    libgtk-4         libgtkmm-4       libgtksourceview5   libheif            libpng            libportal              libportal-gtk4      libreoffice-common     librsvg                libsodium                   libtiff
            libtool      libvte           libwebkit-gtk41  libwebkit-gtk6      lzo                mesalib           miniupnpc              nautilus-extension  nemo                   network-manager        networkmanager-openconnect  openconnect
            pipewire     pipewire-lib     pixman           poppler             poppler-utils      postgresql-libpq  python-pysmbc          qt6-base            qt6-declarative        qt6-multimedia         qt6-quick3d                 qt6-quicktimeline
            qt6-wayland  rav1e            rhythmbox        samba               sdl2               svt-av1           thunderbird            wayland             xapp                   xmlsec1                xorg-server                 xorg-xwayland
            xreader      xviewer          zenity
            Total size of package(s): 1.94 GB
            Downloading 1 / 111
            Package ldb found in repository Solus
            ldb-2.8.2-31-1-x86_64.eopkg    (137.0 KB)100%      0.00 --/- [--:--:--] [complete]
            (...)
            Package 0ad-data found in repository Solus
            0ad-data-0.0.26a-10-1-x86_64.eopkg (1.4 GB) 39%
            (...)
        """
        return self.run_cli("install", "--yes-all", package_id, sudo=True)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ sudo eopkg --no-color upgrade --yes-all
        """
        return self.build_cli("upgrade", sudo=True)

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ sudo eopkg --no-color upgrade --yes-all xz
            Updating repositories
            Updating repository: Solus
            eopkg-index.xml.xz.sha1sum     (40.0  B)100%      0.00 --/- [--:--:--] [complete]
            Solus repository information is up-to-date.
            Warning: Safety switch forces the installation of following packages:
            os-release
            Warning: Safety switch forces the upgrade of following packages:
            bash    bash-completion  brotli   eopkg    gawk            glib2    glibc         gobject-introspection  hwdata  json-c   libcap2  libdw
            libelf  libjson-glib     libnspr  libnss   libpipeline     libssh2  libunistring  lvm2                   lzip    ncurses  nghttp2  nghttp3
            pisi    readline         sqlite3  systemd  wireless-regdb  xz
            Total size of package(s): 55.40 MB
            Warning: There are extra packages due to dependencies.
            Downloading 1 / 32
            Package ncurses found in repository Solus
            ncurses-6.5.20241006-29-1-x86_64.eopkg (767.0 KB)100%      0.00 --/- [--:--:--] [complete]
            (...)
            [✓] Syncing filesystems                                                success
            [✓] Updating dynamic library cache                                     success
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [✓] Updating hwdb                                                      success
            [✓] Updating system users                                              success
            [✓] Updating systemd tmpfiles                                          success
            [✓] Reloading systemd configuration                                    success
            [ ] Re-starting vendor-enabled .socket units                           skipped
            [ ] Re-executing systemd                                               skipped
            [✓] Compiling glib-schemas                                             success
            [✓] Creating GIO modules cache                                         success
            [✓] Updating manpages database                                         success
            [✓] Reloading udev rules                                               success
            [✓] Applying udev rules                                                success
        """
        return self.build_cli("upgrade", "--yes-all", package_id, sudo=True)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ sudo eopkg --no-color remove --yes-all firefox
            The following list of packages will be removed
            in the respective order to satisfy dependencies:
            firefox
            Removing package firefox
            Rebuilding the FilesDB...
            Adding packages to FilesDB /var/lib/eopkg/info/files.db:
            ................
            847 packages added in total.
            Done rebuilding FilesDB (version: 3)
            Removed firefox
            [✓] Syncing filesystems                                                success
            [✓] Updating dynamic library cache                                     success
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Updating clr-boot-manager                                          skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Registering QoL migration on next boot                             skipped
            [ ] Re-starting vendor-enabled .socket units                           skipped
            [ ] Re-executing systemd                                               skipped
            [✓] Updating icon theme cache: hicolor                                 success
            [✓] Updating desktop database                                          success
            [✓] Updating manpages database                                         success
        """
        return self.run_cli("remove", "--yes-all", package_id, sudo=True)

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            $ sudo --no-color eopkg update-repo
            Updating repository: Solus
            eopkg-index.xml.xz.sha1sum  (40.0  B)100%   0.00 --/- [--:--:--] [complete]
            eopkg-index.xml.xz           (3.1 MB)100%  87.40 KB/s [00:00:34] [complete]
            Package database updated.
        """
        self.run_cli("update-repo", sudo=True)

    def cleanup(self) -> None:
        """Removes things we don't need anymore:
        - orphaned packages,
        - outdated package locks
        - package cache and package manager cache

        .. code-block:: shell-session

            $ sudo eopkg --no-color remove-orphans --yes-all
            $ sudo eopkg --no-color clean
            $ sudo eopkg --no-color delete-cache
        """
        self.run_cli("remove-orphans", "--yes-all", sudo=True)
        self.run_cli("clean", sudo=True)
        self.run_cli("delete-cache", sudo=True)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import json
import re
from functools import cached_property

from extra_platforms import ALL_PLATFORMS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package
    from ..version import TokenizedString


class Pip(PackageManager):
    """We will use the Python binary to call out ``pip`` as a module instead of a CLI.

    This is a more robust way of managing packages: "if you're on Windows there
    is an added benefit to using `python -m pip` as it lets `pip` update itself."
    Source: https://snarky.ca/why-you-should-use-python-m-pip/
    """

    homepage_url = "https://pip.pypa.io"

    platforms = ALL_PLATFORMS

    requirement = "10.0.0"

    _SEARCH_REGEXP = re.compile(
        r"""
        ^(?P<package_id>\S+)  # A string with a char at least.
        \                     # A space.
        \((?P<version>.*?)\)  # Content between parenthesis.
        [ ]+-                 # A space or more, then a dash.
        (?P<description>      # Start of the multi-line desc group.
            (?:[ ]+.*\s)+     # Lines of strings prefixed by spaces.
        )
        """,
        re.MULTILINE | re.VERBOSE,
    )

    # Targets `python3` CLI first to allow for some systems (like macOS) to keep the
    # default `python` CLI tied to the Python 2.x ecosystem.
    cli_names = ("python3", "python")

    pre_args = (
        "-m",
        "pip",  # Canonical call to Python's pip module.
        "--no-color",  # Suppress colored output.
    )

    version_cli_options = (*pre_args, "--version")
    version_regexes = (r"pip\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ python -m pip --no-color --version
        pip 2.0.2 from /usr/local/lib/python/site-packages/pip (python 3.7)
    """

    @cached_property
    def version(self) -> TokenizedString | None:
        """Print Python's own version before Pip's.

        This gives much more context to the user about the environment when a Python
        executable is found but Pip is not.

        Runs:

            .. code-block:: shell-session
                $ python --version --version
                Python 3.10.10 (Feb  8 2023, 05:34) [Clang 14.0.0 (clang-1400.0.29.202)]
        """
        if self.executable:
            self.run_cli(
                ("--version", "--version"),
                auto_pre_cmds=False,
                auto_pre_args=False,
                auto_post_args=False,
                force_exec=True,
            )

        # XXX The sentence below gets modernized with `super().version` by ruff.
        # See: https://beta.ruff.rs/docs/rules/#pyupgrade-up
        # But we're explicitly using the old syntax to bypass `cached_property`.
        return super(Pip, self).version  # noqa: UP008

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ python -m pip --no-color list --format=json --verbose --quiet \
            > | jq
            [
             {
                "version": "1.3",
                "name": "backports.functools-lru-cache",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": "pip"
              },
              {
                "version": "0.9999999",
                "name": "html5lib",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": "pip"
              },
              {
                "name": "setuptools",
                "version": "46.0.0",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": ""
              },
              {
                "version": "2.8",
                "name": "Jinja2",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": ""
              },
              (...)
            ]
        """
        # --quiet is required here to silence warning and error messages
        # mangling the JSON content.
        output = self.run_cli("list", "--format=json", "--verbose", "--quiet")

        if output:
            for package in json.loads(output):
                yield self.package(
                    id=package["name"],
                    installed_version=package["version"],
                )

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ python -m pip --no-color list --format=json --outdated \
            > --verbose --quiet | jq
            [
              {
                "latest_filetype": "wheel",
                "version": "0.7.9",
                "name": "alabaster",
                "latest_version": "0.7.10",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": "pip"
              },
              {
                "latest_filetype": "wheel",
                "version": "0.9999999",
                "name": "html5lib",
                "latest_version": "0.999999999",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": "pip"
               },
              {
                "latest_filetype": "wheel",
                "version": "2.8",
                "name": "Jinja2",
                "latest_version": "2.9.5",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": "pip"
               },
              {
                "latest_filetype": "wheel",
                "version": "0.5.3",
                "name": "mccabe",
                "latest_version": "0.6.1",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": "pip"
               },
              {
                "latest_filetype": "wheel",
                "version": "2.2.0",
                "name": "pycodestyle",
                "latest_version": "2.3.1",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": "pip"
               },
              {
                "latest_filetype": "wheel",
                "version": "2.1.3",
                "name": "Pygments",
                "latest_version": "2.2.0",
                "location": "/usr/local/lib/python3.7/site-packages",
                "installer": ""
               }
            ]
        """
        # --quiet is required here to silence warning and error messages
        # mangling the JSON content.
        output = self.run_cli(
            "list",
            "--format=json",
            "--outdated",
            "--verbose",
            "--quiet",
        )

        if output:
            for package in json.loads(output):
                yield self.package(
                    id=package["name"],
                    installed_version=package["version"],
                    latest_version=package["latest_version"],
                )

    @search_capabilities(extended_support=False, exact_support=False)
    def search_xxx_disabled(
        self,
        query: str,
        extended: bool,
        exact: bool,
    ) -> Iterator[Package]:
        """Fetch matching packages.

        .. warning::
            That function was previously named ``search`` but has been renamed
            to make it invisible from the ``mpm`` framework, disabling search
            feature altogether for ``pip``.

            This had to be done has Pip's maintainers disabled the server-side
            API because of unmanageable high-load. See:
            https://github.com/pypa/pip/issues/5216#issuecomment-744605466

        .. caution::
            Search is extended by default. So we returns the best subset of results and
            let :py:meth:`meta_package_manager.base.PackageManager.refiltered_search`
            refine them

        .. code-block:: shell-session

            $ python -m pip --no-color search abc
            ABC (0.0.0)                 - UNKNOWN
            micropython-abc (0.0.1)     - Dummy abc module for MicroPython
            abc1 (1.2.0)                - a list about my think
            abcd (0.3.0)                - AeroGear Build Cli for Digger
            abcyui (1.0.0)              - Sorry ,This is practice!
            astroabc (1.4.2)            - A Python implementation of an
                                          Approximate Bayesian Computation
                                          Sequential Monte Carlo (ABC SMC)
                                          sampler for parameter estimation.
            collective.js.abcjs (1.10)  - UNKNOWN
            cosmo (1.0.5)               - Python ABC sampler
        """
        output = self.run_cli("search", query)

        for package_id, version, description in self._SEARCH_REGEXP.findall(output):
            yield self.package(
                id=package_id,
                description=description,
                latest_version=version,
            )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ python -m pip --no-color install arrow
            Collecting arrow
              Using cached arrow-1.1.1-py3-none-any.whl (60 kB)
            Collecting python-dateutil>=2.7.0
              Using cached python_dateutil-2.8.2-py2.py3-none-any.whl (247 kB)
            Requirement already satisfied: six>=1.5 in python3.9/site-packages (1.16.0)
            Installing collected packages: python-dateutil, arrow
            Successfully installed arrow-1.1.1 python-dateutil-2.8.2
        """
        return self.run_cli("install", package_id)

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade the package provided as parameter.

        .. code-block:: shell-session

            $ python -m pip --no-color install --user --upgrade six
            Collecting six
              Using cached six-1.15.0-py2.py3-none-any.whl (10 kB)
            Installing collected packages: six
              Attempting uninstall: six
                Found existing installation: six 1.14.0
                Uninstalling six-1.14.0:
                  Successfully uninstalled six-1.14.0
            Successfully installed six-1.15.0

        .. note::

            Pip lacks support of a proper full upgrade command. Raising an error let the
            parent class upgrade packages one by one.

            See: https://github.com/pypa/pip/issues/59
        """
        return self.build_cli("install", "--user", "--upgrade", package_id)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ python -m pip --no-color uninstall --yes arrow
        """
        return self.run_cli("uninstall", "--yes", package_id)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import io
import logging
import re
import sys
from datetime import datetime, timezone

from boltons.ecoutils import get_profile
from cyclonedx.model import ExternalReference, ExternalReferenceType, XsUri
from cyclonedx.model.bom import Bom
from cyclonedx.model.component import Component, ComponentType
from cyclonedx.model.contact import OrganizationalEntity
from cyclonedx.model.lifecycle import LifecyclePhase, PredefinedLifecycle
from cyclonedx.output import make_outputter
from cyclonedx.output.json import JsonV1Dot5
from cyclonedx.schema import OutputFormat, SchemaVersion
from cyclonedx.validation import BaseSchemabasedValidator, make_schemabased_validator
from cyclonedx.validation.json import JsonStrictValidator
from extra_platforms import current_platform
from packageurl import PackageURL
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    CreationInfo,
    Document,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    Package as SPDXPackage,
    PackagePurpose,
    Relationship,
    RelationshipType,
)
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.writer.json import json_writer
from spdx_tools.spdx.writer.rdf import rdf_writer
from spdx_tools.spdx.writer.tagvalue import tagvalue_writer
from spdx_tools.spdx.writer.write_utils import convert
from spdx_tools.spdx.writer.xml import xml_writer
from spdx_tools.spdx.writer.yaml import yaml_writer

from . import __version__

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum  # type: ignore[import-not-found]

TYPE_CHECKING = False
if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from .base import Package, PackageManager


class ExportFormat(StrEnum):
    """A user-friendly version of ``spdx_tools.spdx.formats.FileFormat``.

    Map format to user-friendly IDs.
    """

    JSON = "json"
    XML = "xml"
    YAML = "yaml"
    TAG_VALUE = "tag"
    RDF_XML = "rdf"


class SBOM:
    """Utilities shared by all SBOM classes."""

    def __init__(
        self,
        export_format: ExportFormat = ExportFormat.JSON,  # type: ignore[assignment]
    ) -> None:
        """Defaults to JSON export format."""
        logging.debug(f"Set export format to {export_format}")
        self.export_format = export_format

    @staticmethod
    def autodetect_export_format(file_path: Path) -> ExportFormat | None:
        """Better version of ``spdx_tools.spdx.formats.file_name_to_format`` which is
        based on ``Path`` objects and is case-insensitive.

        .. todo:
            Contribute generic autodetection method to Click Extra?
        """
        suffixes = tuple(s.lower() for s in file_path.suffixes[-2:])
        export_format = None
        if suffixes:
            if suffixes == (".rdf", ".xml") or suffixes[-1] == ".rdf":
                export_format = ExportFormat.RDF_XML
            elif suffixes[-1] == ".json":
                export_format = ExportFormat.JSON
            elif suffixes[-1] == ".xml":
                export_format = ExportFormat.XML
            elif suffixes[-1] in (".yaml", ".yml"):
                export_format = ExportFormat.YAML
            elif suffixes[-1] in (".tag", ".spdx"):
                export_format = ExportFormat.TAG_VALUE
        logging.debug(f"File suffixes {suffixes} resolves to {export_format}.")
        return export_format  # type: ignore[return-value]


class SPDX(SBOM):
    """Generates an SPDX document from a list of packages.

    `SPDX 2.3 specifications <https://spdx.github.io/spdx-spec/v2.3/>`_.
    """

    DOC_ID = "SPDXRef-DOCUMENT"
    """Document root ID."""

    document: Document

    @classmethod
    def normalize_spdx_id(cls, str: str) -> str:
        """SPDX IDs must only contain letters, numbers, ``.`` and ``-``."""
        return "-".join(s for s in re.split(r"[^a-zA-Z0-9\.]", str) if s)

    def init_doc(self) -> None:
        """
        `SPDX document metadata specifications
        <https://spdx.github.io/spdx-spec/v2.3/document-creation-information/>`_.
        """
        profile = get_profile()
        system_id = self.normalize_spdx_id(
            "-".join((
                current_platform().name,
                profile["linux_dist_name"],
                profile["linux_dist_version"],
                profile["uname"]["system"],
                profile["uname"]["release"],
                profile["uname"]["machine"],
            ))
        )

        self.document = Document(
            CreationInfo(
                spdx_version="SPDX-2.3",
                spdx_id=self.DOC_ID,
                # Because mpm is a system-wide tool, we chose to name the document
                # after the host operating system platform it was run on.
                name=system_id,
                # Point directly to the mpm release on GitHub so we can get some
                # additional meaning from an URI that is not supposed to have any
                # meaning. We add a trailing "/<random_unique_id>" to the URI as the
                # namespace is supposed to be unique for each document. See:
                # https://spdx.github.io/spdx-spec/v2.3/document-creation-information/#65-spdx-document-namespace-field
                document_namespace=(
                    "https://github.com/kdeldycke/meta-package-manager/releases/tag/"
                    f"v{__version__}/{profile['guid']}"
                ),
                creators=[Actor(ActorType.TOOL, f"meta-package-manager-{__version__}")],
                created=datetime.now(tz=timezone.utc),
                data_license="CC0-1.0",
            )
        )

    def add_package(self, manager: PackageManager, package: Package) -> None:
        """
        `SPDX package metadata specifications
        <https://spdx.github.io/spdx-spec/v2.3/package-information/>`_.
        """
        # pURL string, by its virtue of containing all important metadata of a package,
        # makes perfect unique IDs.
        package_docid = self.normalize_spdx_id(f"SPDXRef-{package.purl}")
        self.document.packages.append(
            SPDXPackage(
                name=package.id,
                spdx_id=package_docid,
                version=str(package.installed_version),
                supplier=Actor(ActorType.ORGANIZATION, manager.name),
                # TODO: Use real URL.
                download_location="https://www.example.com",
                # Current SPDX export is only compiling metadata about packages and their
                # dependencies. So we do not analyze files, as specified in:
                # https://spdx.github.io/spdx-spec/v2.3/package-information/#782-intent
                files_analyzed=False,
                summary=package.name,
                description=package.description,
                external_references=[
                    ExternalPackageRef(
                        ExternalPackageRefCategory.PACKAGE_MANAGER,
                        "purl",
                        package.purl.to_string(),
                    )
                ],
                primary_package_purpose=PackagePurpose.INSTALL,
            )
        )

        # A DESCRIBES relationship asserts that the document indeed describes the
        # package.
        self.document.relationships.append(
            Relationship(self.DOC_ID, RelationshipType.DESCRIBES, package_docid)
        )

    def export(self) -> str:
        """Similar to ``spdx_tools.spdx.writer.write_anything.write_file`` but write
        directly to provided stream instead of file path.
        """
        stream = io.StringIO()

        writer: Any
        if self.export_format == ExportFormat.JSON:
            writer = json_writer
        elif self.export_format == ExportFormat.XML:
            writer = xml_writer
        elif self.export_format == ExportFormat.YAML:
            writer = yaml_writer
        elif self.export_format == ExportFormat.TAG_VALUE:
            writer = tagvalue_writer
        elif self.export_format == ExportFormat.RDF_XML:
            writer = rdf_writer
            # RDF writer expects a binary-mode IO stream.
            stream = io.BytesIO()  # type: ignore[assignment]
        else:
            raise ValueError(f"{self.export_format} not supported.")

        logging.debug("Validate document...")
        errors = validate_full_spdx_document(self.document)
        if errors:
            document_dict = convert(self.document, None)  # type: ignore[arg-type]
            logging.debug(document_dict)
            raise ValueError(f"Document is not valid. Errors: {errors}")

        logging.debug(f"Export with {writer.__name__}")
        writer.write_document_to_stream(self.document, stream, validate=False)
        return stream.getvalue()


class CycloneDX(SBOM):
    """Generates a CycloneDX document from a list of packages.

    `CycloneDX 1.5 specifications <https://cyclonedx.org/docs/1.5>`_.
    """

    document: Bom

    def init_doc(self) -> None:
        """
        `CycloneDX document metadata specifications
        <https://cyclonedx.org/docs/1.5/json/#metadata/>`_.
        """
        gh_url = "https://github.com/kdeldycke/meta-package-manager"
        doc_url = "https://kdeldycke.github.io/meta-package-manager"
        self.document = Bom()

        # mpm produces an inventory of what is installed on a live system.
        self.document.metadata.lifecycles = [
            PredefinedLifecycle(phase=LifecyclePhase.OPERATIONS),
        ]

        self.document.metadata.component = Component(
            name="meta-package-manager",
            type=ComponentType.APPLICATION,
            bom_ref=f"meta-package-manager@{__version__}",
            supplier=OrganizationalEntity(
                name="Meta Package Manager",
                urls=[XsUri(gh_url)],
            ),
            version=__version__,
            purl=PackageURL(
                type="pypi", name="meta-package-manager", version=__version__
            ),
            external_references=[
                ExternalReference(
                    type=ExternalReferenceType.ADVISORIES,
                    url=XsUri(f"{gh_url}/security"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.BUILD_META,
                    url=XsUri(f"{gh_url}/blob/v{__version__}/uv.lock"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.BUILD_SYSTEM,
                    url=XsUri(f"{gh_url}/actions"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.CONFIGURATION,
                    url=XsUri(f"{doc_url}/configuration.html"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.DISTRIBUTION,
                    url=XsUri("https://pypi.org/project/meta-package-manager"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.DISTRIBUTION_INTAKE,
                    url=XsUri(f"{gh_url}/releases/tag/v{__version__}"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.DOCUMENTATION,
                    url=XsUri(doc_url),
                ),
                ExternalReference(
                    type=ExternalReferenceType.ISSUE_TRACKER,
                    url=XsUri(f"{gh_url}/issues"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.LICENSE,
                    url=XsUri(f"{gh_url}/blob/main/license"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.MATURITY_REPORT,
                    url=XsUri(f"{gh_url}/pulse"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.RELEASE_NOTES,
                    url=XsUri(f"{doc_url}/changelog.html"),
                ),
                ExternalReference(
                    type=ExternalReferenceType.SOURCE_DISTRIBUTION,
                    url=XsUri(
                        f"{gh_url}/releases/download/v{__version__}"
                        f"/meta_package_manager-{__version__}.tar.gz"
                    ),
                ),
                ExternalReference(
                    type=ExternalReferenceType.VCS,
                    url=XsUri(gh_url),
                ),
                ExternalReference(
                    type=ExternalReferenceType.WEBSITE,
                    url=XsUri(gh_url),
                ),
                ExternalReference(
                    type=ExternalReferenceType.OTHER,
                    url=XsUri("https://github.com/sponsors/kdeldycke"),
                    comment="Funding",
                ),
            ],
        )

    def add_package(self, manager: PackageManager, package: Package) -> None:
        """
        `CycloneDX package metadata specifications
        <https://cyclonedx.org/docs/1.5/json/#components>`_.
        """
        data = Component(
            name=package.id,
            type=ComponentType.APPLICATION,
            # pURL string, by its virtue of containing all important metadata of a
            # package, makes perfect unique IDs.
            bom_ref=package.purl.to_string(),
            group=package.manager_id,
            version=str(package.installed_version),
            description=package.description,
            purl=package.purl,
        )
        self.document.components.add(data)
        self.document.register_dependency(
            self.document.metadata.component,  # type:ignore[arg-type]
            [data],
        )

    def export(self) -> str:
        validator: BaseSchemabasedValidator
        if self.export_format == ExportFormat.JSON:
            content = JsonV1Dot5(self.document).output_as_string(indent=2)
            validator = JsonStrictValidator(SchemaVersion.V1_5)

        elif self.export_format == ExportFormat.XML:
            writer = make_outputter(self.document, OutputFormat.XML, SchemaVersion.V1_6)
            content = writer.output_as_string(indent=2)
            validator = make_schemabased_validator(
                writer.output_format, writer.schema_version
            )

        else:
            raise ValueError(f"{self.export_format} not supported.")

        logging.debug("Validate document...")
        errors = validator.validate_str(content)
        if errors:
            logging.debug(content)
            raise ValueError(f"Document is not valid. Errors: {errors}")

        return content



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import WINDOWS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class Scoop(PackageManager):
    name = "Scoop"

    homepage_url = "https://scoop.sh"

    platforms = WINDOWS

    requirement = "0.2.4"

    _LIST_REGEXP = re.compile(
        r"""
        (?P<package_id>\S+)  # Any string.
        \s+                  # Any number of blank chars.
        (?P<version>\S+)     # Version string.
        \s+                  # Any number of blank chars.
        .+                   # Any string.
        """,
        re.VERBOSE,
    )
    _OUTDATED_REGEXP = re.compile(
        r"""
        (?P<package_id>\S+)         # Any string.
        \ +                         # One space or more.
        (?P<installed_version>\S*)  # Version string.
        \ +                         # One space or more.
        (?P<latest_version>\S*)     # Version string.
        """,
        re.VERBOSE,
    )
    _SEARCH_REGEXP = re.compile(
        r"""
        (?P<package_id>\S+)  # Any string.
        \ +                  # One space or more.
        (?P<version>\S+)     # Version string.
        \ +                  # One space or more.
        \S+                  # Any string.
        """,
        re.VERBOSE,
    )

    version_regexes = (
        r"^v(?P<version>\S+)\s.+",
        # XXX Scoop does not always provide a clean version string:
        # https://github.com/ScoopInstaller/Scoop/issues/6457
        # https://github.com/ScoopInstaller/Scoop/issues/6270
        # https://github.com/ScoopInstaller/Scoop/pull/6463
        r"^.+,\stag:\sv(?P<version>\S+),\s.+",
        r"^.+\sBump\sto\sversion\s(?P<version>\S+)\s.+",
    )
    """Search version at the start of a line.

    .. code-block:: pwsh-session

        > scoop --version
        Current Scoop version:
        v0.2.4 - Released at 2022-08-08

        'main' bucket:
        5a5b13b6c (HEAD -> master, origin/HEAD) oh-my-posh: Update to version 11.1.1

    .. attention::

        `Scoop does not always provide a clean version string
        <https://github.com/ScoopInstaller/Scoop/issues/6457>`_.

        So we fallback on parsing various ``git log`` output:

        .. code-block:: pwsh-session

            > scoop --version
            Current Scoop version:
            b588a06e (HEAD -> master, tag: v0.5.3, origin/master, origin/HEAD) Bump 0.5.3

            'main' bucket:
            46c50c6b0 (HEAD -> master, origin/master, origin/HEAD) fix arm64 version

        .. code-block:: pwsh-session

            > scoop --version
            Current Scoop version:
            b588a06e chore(release): Bump to version 0.5.3 (resync) (#6436)

            'main' bucket:
            46c50c6b0 aqua: fix arm64 version (#7071)
    """

    @staticmethod
    def remove_headers(text: str) -> str:
        results = text.split("---\n", 1)
        if len(results) != 2:
            return ""
        return results[1]

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: pwsh-session

            >  scoop list
            Installed apps:

            Name   Version          Source Updated             Info
            ----   -------          ------ -------             ----
            7zip   22.01            main   2022-09-27 08:03:30
            dark   3.11.2           main   2022-09-27 08:04:26
            git    2.37.3.windows.1 main   2022-09-27 08:03:58
            python 3.10.7           main   2022-09-27 08:04:53
        """
        output = self.run_cli("list")

        for package_id, version in self._LIST_REGEXP.findall(self.remove_headers(output)):
            yield self.package(id=package_id, installed_version=version)

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: pwsh-session

            > scoop status
            Name           Installed Version Latest Version Missing Dependencies Info
            ----           ----------------- -------------- -------------------- ----
            demulshooter   16.7.2            18.7.3
            eduke32        20220611-10112    20220709-10115
            Teracopy-np
            yuzu-pineapple EA-2804           EA-2830
        """
        output = self.run_cli("scoop", "status")

        for package_id, installed_version, latest_version in self._OUTDATED_REGEXP.findall(
            self.remove_headers(output),
        ):
            yield self.package(
                id=package_id,
                latest_version=latest_version,
                installed_version=installed_version,
            )

    @search_capabilities(extended_support=False, exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not support extended or exact matching. So we returns the best
            subset of results and let
            :py:meth:`meta_package_manager.base.PackageManager.refiltered_search` refine
            them.

        .. code-block:: pwsh-session

            > scoop search zip
            Results from local buckets...

            Name             Version         Source Binaries
            ----             -------         ------ --------
            7zip             22.01           main
            7zip19.00-helper 19.00           main
            busybox          4716-g31467ddfc main   bunzip2 | gunzip | gzip | unzip
            bzip2            1.0.8.0         main
            gow              0.8.0           main   bunzip2.exe | bzip2.exe | zip.exe
            gzip             1.3.12          main
            lzip             1.20            main
            unzip            6.00            main
            zip              3.0             main
        """
        output = self.run_cli("search", query)

        for package_id, version in self._SEARCH_REGEXP.findall(
            self.remove_headers(output),
        ):
            yield self.package(id=package_id, latest_version=version)

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: pwsh-session

            > scoop install 7zip
            Installing '7zip' (22.01) [64bit] from main bucket
            7z2201-x64.msi (1.8 MB) [====================] 100%
            Checking hash of 7z2201-x64.msi ... ok.
            Extracting 7z2201-x64.msi ... done.
            Linking ~\\scoop\apps\7zip\\current => ~\\scoop\apps\7zip\22.01
            Creating shim for '7z'.
            Creating shortcut for 7-Zip (7zFM.exe)
            Persisting Codecs
            Persisting Formats
            Running post_install script...
            '7zip' (22.01) was installed successfully!

            Notes
            -----
            Add 7-Zip as a context menu by running: "C:\\scoop\\...install-context.reg"
        """
        return self.run_cli("install", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: pwsh-session

            > scoop update --all
        """
        return self.build_cli("update", "--all")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: pwsh-session

            > scoop update 7zip
        """
        return self.build_cli("update", package_id)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: pwsh-session

            > scoop uninstall 7zip --purge
            Uninstalling '7zip' (22.01).
            Removing shim '7z.shim'.
            Removing shim '7z.exe'.
            Removing shortcut ~\\AppData\\Roaming\\Scoop Apps\7-Zip.lnk
            Unlinking ~\\scoop\apps\7zip\\current
            '7zip' was uninstalled.
        """
        return self.run_cli("uninstall", package_id, "--purge")

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: pwsh-session

            > scoop status
            WARN  Scoop out of date. Run 'scoop update' to get the latest changes.

        .. code-block:: pwsh-session

            > scoop update
            Updating Scoop...
            Updating 'main' bucket...
            Converting 'main' bucket to git repo...
            Checking repo... OK
            The main bucket was added successfully.
            Scoop was updated successfully!

        .. code-block:: pwsh-session

            > scoop status
            Scoop is up to date.
            Everything is ok!
        """
        self.run_cli("update")

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        .. code-block:: pwsh-session

            > scoop cleanup --all --cache
            Everything is shiny now!
        """
        self.run_cli("cleanup", "--all", "--cache")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import json
import re

from extra_platforms import ALL_PLATFORMS

from ..base import Package, PackageManager

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class UVBase(PackageManager):
    """Virtual package manager shared by ``UV`` and ``UVX`` CLI defined below.

    .. hint::
        Package spec are not quoted anywhere here to `workaround issues
        <https://github.com/kdeldycke/meta-package-manager/issues/1653>`_ with how `uv`
        fails to parses them sometimes.
    """

    homepage_url = "https://docs.astral.sh/uv"

    requirement = "0.5.0"
    """`0.5.0 <https://github.com/astral-sh/uv/releases/tag/0.5.0>`_ is the first
    version to introduce ``pip list --outdated`` command.
    """

    platforms = ALL_PLATFORMS

    # Declare this manager as virtual, i.e. not tied to a real CLI.
    virtual = True

    pre_args = ("--color", "never", "--no-progress")
    """
    - ``--color color-choice``
        Control colors in output [default: ``auto``]

        Possible values:
        - ``auto``: Enables colored output only when the output is going to a terminal or TTY with support
        - ``always``: Enables colored output regardless of the detected environment
        - ``never``: Disables colored output

    - ``--no-progress``
        Hide all progress outputs.

        For example, spinners or progress bars.
    """

    version_regexes = (r"uv\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ uv --version
        uv 0.2.21 (ebfe6d8fc 2024-07-03)
    """

    def _build_package_spec(self, package_id: str, version: str | None = None) -> str:
        """Build package specification with optional version constraint."""
        package_specs = package_id
        if version:
            package_specs += f"=={version}"
        return package_specs


class UV(UVBase):
    """Python package manager using uv pip commands."""

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ uv --color never --no-progress pip list --format=json | jq
            [
              {
                "name": "markupsafe",
                "version": "2.1.5"
              },
              {
                "name": "meta-package-manager",
                "version": "5.17.0",
                "editable_project_location": "/Users/kde/meta-package-manager"
              },
              {
                "name": "myst-parser",
                "version": "3.0.1"
              },
              (...)
            ]
        """
        output = self.run_cli("pip", "list", "--format=json")

        if output:
            for package in json.loads(output):
                yield self.package(
                    id=package["name"],
                    installed_version=package["version"],
                )

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ uv --color never --no-progress pip list --outdated --format=json | jq
            [
              {
                "name": "lark-parser",
                "version": "0.7.8",
                "latest_version": "0.12.0",
                "latest_filetype": "wheel"
              },
              {
                "name": "types-setuptools",
                "version": "75.3.0.20241107",
                "latest_version": "75.3.0.20241112",
                "latest_filetype": "wheel"
              }
            ]
        """
        output = self.run_cli("pip", "list", "--outdated", "--format=json")

        if output:
            for package in json.loads(output):
                yield self.package(
                    id=package["name"],
                    installed_version=package["version"],
                    latest_version=package["latest_version"],
                )

    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ uv --color never --no-progress pip install arrow
        """
        package_specs = self._build_package_spec(package_id, version)
        return self.run_cli("pip", "install", package_specs)

    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade the package provided as parameter.

        .. code-block:: shell-session

            $ uv --color never --no-progress pip install --upgrade arrow
        """
        package_specs = self._build_package_spec(package_id, version)
        return self.build_cli("pip", "install", "--upgrade", package_specs)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ uv --color never --no-progress pip uninstall arrow
        """
        return self.run_cli("pip", "uninstall", package_id)

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        .. code-block:: shell-session

            $ uv --color never --no-progress cache clean
            Clearing cache at: /Users/kde/Library/Caches/uv
            Removed 97279 files (2.0GiB)

        .. code-block:: shell-session

            $ uv --color never --no-progress cache prune
            No cache found at: /Users/kde/.cache/uv
        """
        self.run_cli("cache", "clean")
        self.run_cli("cache", "prune")


class UVX(UVBase):
    """UV tool manager for isolated Python tools.

    Like ``pipx``, but uses ``uv tool`` commands.
    """

    cli_names = ("uv",)

    homepage_url = "https://docs.astral.sh/uv/guides/tools/"

    _INSTALLED_REGEXP = re.compile(r"^(?P<package_id>\S+)\s+v(?P<version>\S+)$")

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ uv --color never --no-progress tool list
            pycowsay v0.0.0.1
            - pycowsay
        """
        output = self.run_cli("tool", "list")

        if output:
            for line in output.splitlines():
                match = self._INSTALLED_REGEXP.match(line)
                if match:
                    yield self.package(
                        id=match.group("package_id"),
                        installed_version=match.group("version"),
                    )

    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ uv --color never --no-progress tool install pycowsay
        """
        package_specs = self._build_package_spec(package_id, version)
        return self.run_cli("tool", "install", package_specs)

    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade the package provided as parameter.

        .. code-block:: shell-session

            $ uv --color never --no-progress tool upgrade pycowsay
        """
        package_specs = self._build_package_spec(package_id, version)
        return self.build_cli("tool", "upgrade", package_specs)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages.

        .. code-block:: shell-session

            $ uv --color never --no-progress tool upgrade --all
            Updated pycowsay v0.0.0.1 -> v0.0.0.2
                - pycowsay
        """
        return self.build_cli("tool", "upgrade", "--all")

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ uv --color never --no-progress tool uninstall pycowsay
        """
        return self.run_cli("tool", "uninstall", package_id)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

from extra_platforms import WINDOWS

from ..base import PackageManager

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Generator, Iterator

    from ..base import Package


class WinGet(PackageManager):
    homepage_url = "https://github.com/microsoft/winget-cli"

    platforms = WINDOWS

    requirement = "1.7"

    post_args = ("--accept-source-agreements", "--disable-interactivity")
    """
    ``--accept-source-agreements``:
        Used to accept the source license agreement, and avoid the following prompt:

        .. code-block:: pwsh-session
            PS C:\\Users\\kev> winget list
            The "msstore' source requires that you view the following agreements before using.
            Terms of Transaction: https://aka.ms/microsoft-store-terms-of-transaction
            The source requires the current machine's 2-letter geographic region to be sent to the backend service to function prope rly (ex. "US").

            Do you agree to all the source agreements terms?
            [Y] Yes [N] No:

    ``--disable-interactivity``:
        Disable interactive prompts.

    ..todo::
        Add the ``--no-progress`` option once it is available in the stable release:
        - https://github.com/microsoft/winget-cli/pull/6049
        - https://github.com/microsoft/winget-cli/issues/3494#issuecomment-3921618377
    """

    version_regexes = (r"v\s+(?P<version>\S+)",)
    """
    .. code-block:: pwsh-session

        PS C:\\Users\\kev> winget --version
        v1.7.11261
    """

    def _parse_table(self, output: str) -> Iterator[Generator[str, None, None]]:
        """Parse a table from the output of a winget command and returns a generator of cells."""
        # Extract table.
        table_start = "Name "
        assert table_start in output, (
            f"Cannot find table starting with {table_start!r} in:\n{output}"
        )
        assert output.count(table_start) == 1, (
            f"{table_start!r} not unique in:\n{output}"
        )
        table = output.split(table_start, 1)[1]

        # Check table format.
        lines = table.splitlines()
        table_width = len(lines[0])
        assert lines[1] == "-" * table_width, (
            f"Table headers not followed by expected separator:\n{table}"
        )
        assert all(len(line) <= table_width for line in lines[2:]), (
            f"Table lines with different width:\n{table}"
        )

        # Guess column positions.
        headers = []
        col_str = ""
        for char in lines[0]:
            if col_str and char != " " and " " in col_str:
                headers.append(col_str)
                col_str = ""
            col_str += char
        if col_str:
            headers.append(col_str)

        col_ranges = []
        for header in headers:
            start = lines[0].index(header)
            end = start + len(header)
            col_ranges.append((start, end))

        for line in lines[2:]:
            yield (line[start:end].strip() for start, end in col_ranges)

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget list --accept-source-agreements --disable-interactivity
            The 'msstore' source requires that you view the following agreements before using.
            Terms of Transaction: https://aka.ms/microsoft-store-terms-of-transaction
            The source requires the current machine's 2-letter geographic region to be sent to the backend service to function properly (ex. "US").

            Name                          Id                           Version        Available     Source
            ----------------------------------------------------------------------------------------------
            CCleaner                      CCleaner                     6.08
            Git                           Git.Git                      2.37.3         2.45.1        winget
            Microsoft Edge                Microsoft.Edge               109.0.1518.70  125.0.2535.51 winget
            Microsoft Edge Update         Microsoft Edge Update        1.3.187.37
            App Installer                 Microsoft.AppInstaller       1.21.3482.0                  winget
            Microsoft.UI.Xaml.2.7         Microsoft.UI.Xaml.2.7        7.2208.15002.0               winget
            Python Launcher               Python.Launchez              < 3.12.0       3.12.0        winget
            Microsoft Visual C++ (x86)... Microsoft.VCRedist.2015+.X86 14.34.31931.0  14.38.33135.0 winget
        """
        output = self.run_cli("list")

        for name, package_id, installed_version, _, _ in self._parse_table(output):
            # Strip the "<" comparison operator from the version.
            if " " in installed_version:
                installed_version = installed_version.split()[1]

            yield self.package(
                id=package_id, name=name, installed_version=installed_version
            )

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget list --upgrade-available --accept-source-agreements --disable-interactivity
            Name                          Id                           Version       Available     Source
            ---------------------------------------------------------------------------------------------
            Git                           Git.Git                      2.37.3        2.45.1        winget
            Microsoft Edge                Microsoft.Edge               109.0.1518.70 125.0.2535.51 winget
            Python Launcher               Python.Launchez              < 3.12.0      3.12.0        winget
            Microsoft Visual C++ (x86)... Microsoft.VCRedist.2015+.X86 14.34.31931.0 14.38.33135.0 winget
            4 upgrades available.
        """
        output = self.run_cli("list", "--upgrade-available").strip()
        if output.endswith(" upgrades available."):
            output = "\n".join(output.splitlines()[:-1])

        for name, package_id, installed_version, latest_version, _ in self._parse_table(
            output
        ):
            # Strip the "<" comparison operator from the version.
            if " " in installed_version:
                installed_version = installed_version.split()[1]

            yield self.package(
                id=package_id,
                name=name,
                installed_version=installed_version,
                latest_version=latest_version,
            )

    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget search --query vscode --accept-source-agreements --disable-interactivity
            Name                             Id                               Version      Match               Source
            ---------------------------------------------------------------------------------------------------------
            Microsoft Visual Studio Code     Microsoft.VisualStudioCode       1.89.1       Moniker: vscode     winget
            MrCode                           zokugun.MrCode                   1.82.0.23253 Tag: vscode         winget
            VSCodium Insiders                VSCodium.VSCodium.Insiders       1.88.0.24095 Tag: vscode         winget
            VSCodium                         VSCodium.VSCodium                1.89.1.24130 Tag: vscode         winget
            Upgit                            pluveto.Upgit                    0.2.18       Tag: vscode         winget
            vscli                            michidk.vscli                    0.3.0        Tag: vscode         winget
            Huawei QuickApp IDE              Huawei.QuickAppIde               14.0.1       Tag: vscode         winget
            TheiaBlueprint                   EclipseFoundation.TheiaBlueprint 1.44.0       Tag: vscode         winget
            Codium                           Alex313031.Codium                1.86.2.24053 Tag: vscode         winget
            Cursor Editor                    CursorAI,Inc.Cursor              latest       Tag: vscode         winget
            Microsoft Visual Studio Code CLI Microsoft.VisualStudioCode.CLI   1.89.1       Moniker: vscode-cli winget

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget search --query vscode --exact --accept-source-agreements --disable-interactivity
            Name                         Id                               Version      Match           Source
            -------------------------------------------------------------------------------------------------
            Microsoft Visual Studio Code Microsoft.VisualStudioCode       1.89.1       Moniker: vscode winget
            MrCode                       zokugun.MrCode                   1.82.0.23253 Tag: vscode     winget
            VSCodium Insiders            VSCodium.VSCodium.Insiders       1.88.0.24095 Tag: vscode     winget
            VSCodium                     VSCodium.VSCodium                1.89.1.24130 Tag: vscode     winget
            Upgit                        pluveto.Upgit                    0.2.18       Tag: vscode     winget
            vscli                        michidk.vscli                    0.3.0        Tag: vscode     winget
            Huawei QuickApp IDE          Huawei.QuickAppIde               14.0.1       Tag: vscode     winget
            TheiaBlueprint               EclipseFoundation.TheiaBlueprint 1.44.0       Tag: vscode     winget
            Codium                       Alex313031.Codium                1.86.2.24053 Tag: vscode     winget
            Cursor Editor                CursorAI,Inc.Cursor              latest       Tag: vscode     winget

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget search --id VSCodium.VSCodium --accept-source-agreements --disable-interactivity
            Name              Id                         Version      Source
            ----------------------------------------------------------------
            VSCodium Insiders VSCodium.VSCodium.Insiders 1.88.0.24095 winget
            VSCodium          VSCodium.VSCodium          1.89.1.24130 winget

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget search --name Codium --accept-source-agreements --disable-interactivity
            Name              Id                         Version      Source
            ----------------------------------------------------------------
            Codium            Alex313031.Codium          1.86.2.24053 winget
            VSCodium Insiders VSCodium.VSCodium.Insiders 1.88.0.24095 winget
            VSCodium          VSCodium.VSCodium          1.89.1.24130 winget

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget search --id VSCodium.VSCodium  --exact --accept-source-agreements --disable-interactivity
            Name     Id                Version      Source
            ----------------------------------------------
            VSCodium VSCodium.VSCodium 1.89.1.24130 winget

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget search --name Codium --exact --accept-source-agreements --disable-interactivity
            Name   Id                Version      Source
            --------------------------------------------
            Codium Alex313031.Codium 1.86.2.24053 winget
        """
        # Default search is extended to all metadata: id, name, moniker and tag.
        if extended:
            args = ["search", "--query", query]
            # Exact search deactivates substring search.
            if exact:
                args.append("--exact")
            output = self.run_cli(args)
            for name, package_id, version, _, _ in self._parse_table(output):
                yield self.package(id=package_id, name=name, latest_version=version)

        # For non-extended search, we need to perform 2 queries, one for id and
        # one for name.
        else:
            for field in "--id", "--name":
                output = self.run_cli(
                    "search", field, query, "--exact" if exact else None
                )
                for name, package_id, version, _ in self._parse_table(output):
                    yield self.package(id=package_id, name=name, latest_version=version)

    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget install --id Microsoft.PowerToys --accept-package-agreements --version 0.15.2 --accept-source-agreements --disable-interactivity
            Found Power Toys [Microsoft.PowerToys] Version 0.15.2
            This application is licensed to you by its owner.
            Microsoft is not responsible for, nor does it grant any licenses to, third-party packages.
            Successfully verified installer hash
            Starting package install...
              ██████████████████████████████  100%
            Successfully installed
        """
        args = ["install", "--id", package_id, "--accept-package-agreements"]
        if version:
            args += ["--version", version]
        return self.run_cli(args)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget upgrade --all --accept-package-agreements --accept-source-agreements --disable-interactivity
            Name                            Id                            Version       Available     Source
            ------------------------------------------------------------------------------------------------
            Microsoft Edge                  Microsoft.Edge                109.0.1518.70 125.0.2535.51 winget
            Microsoft Edge WebView2 Runtime Microsoft.EdgeWebView2Runtime 109.0.1518.70 125.0.2535.51 winget
            Python Launcher                 Python.Launchez               < 3.12.0      3.12.0        winget
            Microsoft Visual C++ (x86)...   Microsoft.VCRedist.2015+.X86  14.34.31931.0 14.38.33135.0 winget
            4 upgrades available.

            Installing dependencies:
            This package requires the following dependencies:
              - Packages
                  Microsoft.UI.Xaml.2.8 [>= 8.2306.22001.0]
            (1/3) Found Microsoft Edge WebView2 Runtime [Microsoft.EdgeWebView2Runtime] Version 125. 0.2535.51
            This application is licensed to you by its owner.
            Microsoft is not responsible for, nor does it grant any licenses to, third-party packages.
            Downloading https://msedge.sf.dl.delivery.mp.microsoft.com/filestreamingservice/files/e5dd841e-17ff-43b7-a2c0-ff759f55c202/MicrosoftEdgeWebView2RuntimeInstallerARM64.exe
              ██████████████████████████████  166 MB /  166 MB
            Successfully verified installer hash
            Starting package install...
            Successfully installed

            (...)
        """
        return self.build_cli("update", "--all", "--accept-package-agreements")

    def upgrade_one_cli(
        self, package_id: str, version: str | None = None
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget upgrade --id Git.Git --accept-package-agreements --accept-source-agreements --disable-interactivity
            Found Git [Git.Git] Version 2.45.1
            This application is licensed to you by its owner.
            Microsoft is not responsible for, nor does it grant any licenses to, third-party packages.
            Downloading https://github.com/git-for-windows/git/releases/download/v2.45.1.windows.1/Git-2.45.1-64-bit.exe
              ██████████████████████████████  64.7 MB / 64.7 MB
            Successfully verified installer hash
            Starting package install...
            Successfully installed

        .. todo::

            Automatically uninstall the package if the technology is different:

            .. code-block:: pwsh-session

                PS C:\\Users\\kev> winget upgrade --id Microsoft.Edge
                A newer version was found, but the install technology is different from the current version installed. Please uninstall the package and install the newer version.
        """
        args = ["install", "--id", package_id, "--accept-package-agreements"]
        if version:
            args += ["--version", version]
        return self.build_cli(args)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: pwsh-session

            PS C:\\Users\\kev> winget uninstall --id Microsoft.PowerToys --source winget --accept-source-agreements --disable-interactivity
            Found PowerToys (Preview) [Microsoft.PowerToys]
            Starting package uninstall...
              ██████████████████████████████  100%
            Successfully uninstalled
        """
        return self.run_cli("uninstall", "--id", package_id, "--source", "winget")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import UNIX_WITHOUT_MACOS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class Snap(PackageManager):
    homepage_url = "https://snapcraft.io"

    platforms = UNIX_WITHOUT_MACOS

    requirement = "2.0.0"

    post_args = ("--color=never",)

    _SEARCH_REGEXP = re.compile(
        r"^(?P<package_id>\S+)\s+(?P<version>\S+)\s+\S+\s+\S+\s+(?P<description>.+)$",
        re.MULTILINE,
    )

    version_regexes = (r"snap\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ snap --version
        snap       2.44.1
        snapd      2.44.1
        series     16
        linuxmint  19.3
        kernel     4.15.0-91-generic
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ snap list --color=never
            Name    Version    Rev   Aufzeichnung   Herausgeber     Hinweise
            core    16-2.44.1  8935  latest/stable  canonical✓      core
            wechat  2.0        7     latest/stable  ubuntu-dawndiy  -
            pdftk   2.02-4     9     latest/stable  smoser          -
        """
        output = self.run_cli("list")

        for package in output.splitlines()[1:]:
            parts = package.split()
            yield self.package(id=parts[0], installed_version=parts[1])

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ snap refresh --list --color=never
            Name            Version  Rev  Herausgeber     Hinweise
            standard-notes  3.3.5    8    standardnotes✓  -
        """
        output = self.run_cli("refresh", "--list")

        # Build a lookup of installed versions to avoid one CLI call per package.
        installed_versions = {p.id: p.installed_version for p in self.installed}

        for package in output.splitlines()[1:]:
            parts = package.split()
            package_id = parts[0]
            yield self.package(
                id=package_id,
                latest_version=parts[1],
                installed_version=installed_versions.get(package_id),
            )

    @search_capabilities(extended_support=False, exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search is extended by default. So we returns the best subset of results and
            let :py:meth:`meta_package_manager.base.PackageManager.refiltered_search`
            refine them.

        .. code-block:: shell-session

            $ snap find doc --color=never
            Name       Version      Herausgeber  Hinweise  Zusammenfassung
            journey    2.14.3       2appstudio   -         Your private diary.
            nextcloud  17.0.5snap1  nextcloud✓   -         Nextcloud Server
            skype      8.58.0.93    skype✓       classic   One Skype for all.
        """
        output = self.run_cli("find", query)
        headerless_table = None
        if output:
            table = output.split("\n", 1)
            if len(table) > 1:
                headerless_table = table[1]

        if headerless_table:
            for package_id, version, description in self._SEARCH_REGEXP.findall(
                headerless_table,
            ):
                yield self.package(
                    id=package_id,
                    description=description,
                    latest_version=version,
                )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ snap install standard-notes --color=never
        """
        return self.run_cli("install", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ snap refresh --color=never
        """
        return self.build_cli("refresh")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ snap refresh standard-notes --color=never
        """
        return self.build_cli("refresh", package_id)



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Introspection utilities to produce feature inventory of all managers."""

from __future__ import annotations

from extra_platforms import (
    ALL_WINDOWS,
    BSD_WITHOUT_MACOS,
    LINUX_LIKE,
    MACOS,
    UNIX_WITHOUT_MACOS,
    Group,
)

TYPE_CHECKING = False
if TYPE_CHECKING:
    from extra_platforms import Platform


MAIN_PLATFORMS: tuple[Group | Platform, ...] = (
    BSD_WITHOUT_MACOS.copy(id="bsd", name="BSD"),
    LINUX_LIKE.copy(id="linux", name="Linux", icon="🐧"),
    MACOS,
    UNIX_WITHOUT_MACOS.copy(
        id="unix",
        name="Unix",
        members=tuple(UNIX_WITHOUT_MACOS - BSD_WITHOUT_MACOS - LINUX_LIKE),
    ),
    ALL_WINDOWS.copy(id="windows", name="Windows"),
)
"""Top-level classification of platforms.

This is the local reference used to classify the execution targets of ``mpm``.

Each entry of this list will have its own dedicated column in the matrix. This list is
manually maintained with tweaked IDs and names to minimize the matrix verbosity and
make it readable both in CLI and documentation.

The order of this list determine the order of the resulting columns.
"""



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import MACOS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class MAS(PackageManager):
    name = "Mac AppStore"

    homepage_url = "https://github.com/argon/mas"

    platforms = MACOS

    _INSTALLED_REGEXP = re.compile(
        r"""
        (?P<package_id>\d+)
        \s+
        (?P<package_name>.+?)
        \s+
        \(
            (?P<version>\S+)
        \)
        """,
        re.MULTILINE | re.VERBOSE,
    )

    _OUTDATED_REGEXP = re.compile(
        r"""
        (?P<package_id>\d+)
        \s+
        (?P<package_name>.+?)
        \s+
        \(
            (?P<installed_version>\S+)
            \s+->\s+
            (?P<latest_version>\S+)
        \)
        """,
        re.MULTILINE | re.VERBOSE,
    )

    requirement = "1.8.7"
    """`1.8.7 <https://github.com/mas-cli/mas/releases/tag/v1.8.7>`_ is fixing the
    ``mas search`` command.
    """

    version_cli_options = ("version",)
    """
    .. code-block:: shell-session

        $ mas version
        1.8.3
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ mas list
            1569813296  1Password for Safari                 (2.3.5)
            1295203466  Microsoft Remote Desktop             (10.7.6)
            409183694   Keynote                              (12.0)
            1408727408  com.adriangranados.wifiexplorerlite  (1.5.5)
            409203825   Numbers                              (12.0)
        """
        output = self.run_cli("list")

        for package_id, package_name, version in self._INSTALLED_REGEXP.findall(output):
            yield self.package(
                id=package_id,
                name=package_name,
                installed_version=version,
            )

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ mas outdated
            409183694  Keynote (11.0 -> 12.0)
            1176895641 Spark   (2.11.20 -> 2.11.21)
        """
        output = self.run_cli("outdated")

        for (
            package_id,
            package_name,
            installed_version,
            latest_version,
        ) in self._OUTDATED_REGEXP.findall(output):
            yield self.package(
                id=package_id,
                name=package_name,
                installed_version=installed_version,
                latest_version=latest_version,
            )

    @search_capabilities(extended_support=False, exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. caution::
            Search does not support extended or exact matching. So we returns the best
            subset of results and let
            :py:meth:`meta_package_manager.base.PackageManager.refiltered_search` refine
            them.

        .. code-block:: shell-session

            $ mas search python
               689176796  Python Runner   (1.3)
               630736088  Learning Python (1.0)
               945397020  Run Python      (1.0)
              1164498373  PythonGames     (1.0)
              1400050251  Pythonic        (1.0.0)
        """
        output = self.run_cli("search", query)

        for package_id, package_name, version in self._INSTALLED_REGEXP.findall(output):
            yield self.package(id=package_id, name=package_name, latest_version=version)

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ mas install 945397020
        """
        return self.run_cli("install", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ mas upgrade
        """
        return self.build_cli("upgrade")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ mas upgrade 945397020
        """
        return self.build_cli("upgrade", package_id)

    def remove(self, package_id: str) -> str:
        """Removes a package.

        .. code-block:: shell-session

            $ sudo mas uninstall 1494051017
            Password:
            Deleted '/Applications/SimpleLogin.app' to '/Users/kde/.Trash/SimpleLogin.app'
        """
        return self.run_cli("uninstall", package_id, sudo=True)



#!/usr/bin/env python3
# <xbar.title>Meta Package Manager</xbar.title>
# <xbar.version>v6.1.2.dev0</xbar.version>
# <xbar.author>Kevin Deldycke</xbar.author>
# <xbar.author.github>kdeldycke</xbar.author.github>
# <xbar.desc>List outdated packages and manage upgrades.</xbar.desc>
# <xbar.dependencies>python,mpm</xbar.dependencies>
# <xbar.image>https://raw.githubusercontent.com/kdeldycke/meta-package-manager/refs/heads/main/docs/assets/xbar-submenu-table-rendering.png</xbar.image>
# <xbar.abouturl>https://kdeldycke.github.io/meta-package-manager/bar-plugin.html</xbar.abouturl>
# <xbar.var>boolean(VAR_SUBMENU_LAYOUT=false): Group packages into a sub-menu for each manager.</xbar.var>
# <xbar.var>boolean(VAR_TABLE_RENDERING=true): Aligns package names and versions in a table for easier visual parsing.</xbar.var>
# XXX Deactivate font-related options for Xbar. Default variable value does not allow `=` character in Xbar. See: https://github.com/matryer/xbar/issues/832
# <!--xbar.var>string(VAR_DEFAULT_FONT=""): Font parameters for regular text.</xbar.var-->
# <!--xbar.var>string(VAR_MONOSPACE_FONT="font=Menlo size=12"): Font parameters for monospace text. Used for table rendering and error messages.</xbar.var-->
# <swiftbar.environment>[VAR_SUBMENU_LAYOUT: false, VAR_TABLE_RENDERING: true, VAR_DEFAULT_FONT: , VAR_MONOSPACE_FONT: font=Menlo size=12]</swiftbar.environment>
"""Xbar and SwiftBar plugin for Meta Package Manager (i.e. the :command:`mpm` CLI).

Default update cycle should be set to several hours so we have a chance to get
user's attention once a day. Higher frequency might ruin the system as all
checks are quite resource intensive, and Homebrew might hit GitHub's API calls
quota.

- `Xbar automatically bridge plugin options
  <https://xbarapp.com/docs/2021/03/14/variables-in-xbar.html>`_ between its UI
  and environment variable on script execution.

- This is `in progress for SwiftBar
  <https://github.com/swiftbar/SwiftBar/issues/160>`_.
"""

from __future__ import annotations

import argparse
import os
import re
import sys
from configparser import RawConfigParser
from enum import Enum
from functools import cached_property
from operator import itemgetter, methodcaller
from pathlib import Path
from shlex import shlex
from shutil import which
from subprocess import run
from textwrap import dedent

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Generator


PYTHON_MIN_VERSION = (3, 9, 0)
"""Minimal requirement is aligned to macOS default.

See: https://kdeldycke.github.io/meta-package-manager/bar-plugin.html#python-3-9-required
"""

SWIFTBAR_MIN_VERSION = (2, 1, 2)
"""SwiftBar v2.1.2 fix an issue with multiple parameters in the font strings.

See: https://github.com/swiftbar/SwiftBar/issues/445
"""

XBAR_MIN_VERSION = (2, 1, 7)
"""Xbar v2.1.7-beta is the latest version available on Homebrew."""

MPM_MIN_VERSION = (5, 0, 0)
"""Mpm v5.0.0 was the first version taking care of the complete layout rendering."""

Venv = Enum("Venv", ["PIPENV", "UV", "POETRY", "VIRTUALENV"])
"""Type of virtualenv we are capable of detecting."""


class MPMPlugin:
    """Implements the minimal code necessary to locate and call the ``mpm`` CLI on the
    system.

    Once ``mpm`` is located, we can rely on it to produce the main output of the plugin.

    The output must supports both `Xbar dialect
    <https://github.com/matryer/xbar-plugins/blob/main/CONTRIBUTING.md#plugin-api>`_
    and `SwiftBar dialect <https://github.com/swiftbar/SwiftBar#plugin-api>`_.
    """

    @staticmethod
    def getenv_str(var, default: str | None = None) -> str | None:
        """Utility to get environment variables.

        Note that all environment variables are strings. Always returns a lowered-case
        string.
        """
        value = os.environ.get(var, None)
        if value is None:
            return default
        return str(value).lower()

    @staticmethod
    def getenv_bool(var, default: bool = False) -> bool:
        """Utility to normalize boolean environment variables.

        Relies on ``configparser.RawConfigParser.BOOLEAN_STATES`` to translate strings
        into boolean. See:
        https://github.com/python/cpython/blob/89192c4/Lib/configparser.py#L597-L599
        """
        value = MPMPlugin.getenv_str(var)
        if value is None:
            return default
        return RawConfigParser.BOOLEAN_STATES[value]

    @staticmethod
    def normalize_params(font_string: str, valid_ids: set[str] | None = None) -> str:
        """Parse a multi-parameters string and return a normalized string.

        The string is expected to be a space-separated list of parameters, each
        parameter being a key/value pair separated by an equal sign.

        Only keeps the parameters that are in the ``valid_ids`` set and ignores the
        rest. By default, only ``color``, ``font`` and ``size`` are kept.

        Multiple values for the same parameter will be deduplicated, and the last one
        will be kept.

        Available parameters are:
        - https://github.com/swiftbar/SwiftBar?tab=readme-ov-file#parameters
        - https://github.com/matryer/xbar-plugins/blob/main/CONTRIBUTING.md#parameters
        """
        if not valid_ids:
            valid_ids = {"color", "font", "size"}
        params = {}

        key = None
        previous_token_is_separator = False
        for token in shlex(font_string):
            # Flag the token as a separator if it is an equal sign.
            if token == "=":
                previous_token_is_separator = True
            # Token positioned just after an equal sign is a value. Let's attach it to
            # the key and store it in the params dictionary.
            elif previous_token_is_separator:
                if key and key in valid_ids:
                    params[key] = token
                # Reset the flag and key.
                previous_token_is_separator = False
                key = None
            # Any token is considered a potential key until we find an equal sign.
            else:
                key = token

        return " ".join(f"{k}={v}" for k, v in params.items())

    @staticmethod
    def str_to_version(version_string: str | None) -> tuple[int, ...]:
        """Transforms a string into a tuple of integers representing a version."""
        if not version_string:
            return ()
        return tuple(map(int, version_string.strip().split(".")))

    @staticmethod
    def version_to_str(version_tuple: tuple[int, ...] | None) -> str:
        """Transforms a tuple of integers representing a version into a string."""
        if not version_tuple:
            return "None"
        return ".".join(map(str, version_tuple))

    @cached_property
    def table_rendering(self) -> bool:
        """Aligns package names and versions, like a table, for easier visual parsing.

        If ``True``, will aligns all items using a fixed-width font.
        """
        return self.getenv_bool("VAR_TABLE_RENDERING", True)

    @cached_property
    def default_font(self) -> str:
        """Make it easier to change font, sizes and colors of the output."""
        return self.normalize_params(
            self.getenv_str("VAR_DEFAULT_FONT", ""),  # type: ignore
        )

    @cached_property
    def monospace_font(self) -> str:
        """Make it easier to change font, sizes and colors of the output."""
        return self.normalize_params(
            self.getenv_str("VAR_MONOSPACE_FONT", "font=Menlo size=12"),  # type: ignore
        )

    @cached_property
    def error_font(self) -> str:
        """Error font is based on monospace font."""
        return self.normalize_params(f"{self.monospace_font} color=red size=10")

    @cached_property
    def is_swiftbar(self) -> bool:
        """SwiftBar is kind enough to tell us about its presence."""
        return self.getenv_bool("SWIFTBAR")

    @cached_property
    def all_pythons(self) -> list[str]:
        """Search for any Python on the system.

        Returns a generator of normalized and deduplicated ``Path`` to Python binaries.

        Filters out old Python interpreters.

        We first try to locate Python by respecting the environment variables as-is,
        i.e. as defined by the user. Then we return the Python interpreter used to
        execute this script.

        TODO: try to tweak the env vars to look for homebrew location etc?
        """
        collected = []
        seen = set()
        for bin_name in ("python3", "python", sys.executable):
            py_path = which(bin_name)
            if not py_path:
                continue

            normalized_path = os.path.normcase(Path(py_path).resolve())
            if normalized_path in seen:
                continue
            seen.add(normalized_path)

            process = run(
                (
                    normalized_path,
                    "-c",
                    "import sys; "
                    "v = sys.version_info; "
                    # Ignore releaselevel and serial component of the version number.
                    "print(f'{v.major}.{v.minor}.{v.micro}')",
                ),
                capture_output=True,
                encoding="utf-8",
                check=False,
            )
            python_version = self.str_to_version(process.stdout)
            # Is Python too old?
            if python_version < PYTHON_MIN_VERSION:
                continue

            collected.append(normalized_path)

        return collected

    @staticmethod
    def search_venv(folder: Path) -> tuple[Venv, tuple[str, ...]] | None:
        """Search for signs of a virtual env in the provided folder.

        Returns the type of the detected venv and CLI arguments that can be used to run
        a command from the virtualenv context.

        Returns ``(None, None)`` if the folder is not a venv.

        Inspired by `autoswitch_virtualenv.plugin.zsh
        <https://github.com/MichaelAquilina/zsh-autoswitch-virtualenv/blob/master/autoswitch_virtualenv.plugin.zsh#L50>`_
        and `uv's get_interpreter_info.py
        https://github.com/astral-sh/uv/blob/f770b25/crates/uv-python/python/get_interpreter_info.py>`_.
        """
        if (folder / "Pipfile").is_file():
            return Venv.PIPENV, (f"PIPENV_PIPFILE='{folder}'", "pipenv", "run", "mpm")

        if (folder / "uv.lock").is_file():
            return Venv.UV, ("uv", "run", "mpm")

        if (folder / "poetry.lock").is_file():
            return Venv.POETRY, ("poetry", "run", "--directory", str(folder), "mpm")

        if (folder / "requirements.txt").is_file() or (folder / "setup.py").is_file():
            return Venv.VIRTUALENV, (
                f"VIRTUAL_ENV='{folder}'",
                "python",
                "-m",
                "meta_package_manager",
            )

        return None

    def search_mpm(self) -> Generator[tuple[str, ...], None, None]:
        """Iterate over possible CLI commands to execute ``mpm``.

        Should be able to produce the full spectrum of alternative commands we can use
        to invoke ``mpm`` over different context.

        The order in which the candidates are returned by this method is conserved by
        the ``ranked_mpm()`` method below.

        We prioritize venv-based findings first, as they're more likely to have all
        dependencies installed and sorted out. They're also our prime candidates in
        unittests.

        Then we search for system-wide installation. And finally Python modules.
        """
        # This script might be itself part of an mpm installation that was deployed in
        # a virtualenv. So walk back the whole folder tree from here in search of a
        # virtualenv.
        for folder in Path(__file__).parents:
            # Stop looking beyond Home.
            if folder == Path.home():
                continue

            venv_found = self.search_venv(folder)
            if not venv_found:
                continue

            yield venv_found[1]

        # Search for an mpm executable in the environment, be it a script or a binary.
        mpm_bin = which("mpm")
        if mpm_bin:
            yield (mpm_bin,)

        # Search for a meta_package_manager package installed in any Python found on
        # the system.
        for python_path in self.all_pythons:
            yield python_path, "-m", "meta_package_manager"

    def check_mpm(
        self, mpm_cli_args: tuple[str, ...]
    ) -> tuple[bool, bool, tuple[int, ...] | None, str | Exception | None]:
        """Test-run mpm execution and extract its version."""
        error: str | Exception | None = None
        try:
            process = run(
                # Output a color-less version just in case the script is not run in a
                # non-interactive shell, or Click/Click-Extra autodetection fails.
                (*mpm_cli_args, "--no-color", "--version"),
                capture_output=True,
                encoding="utf-8",
                check=False,
            )
            error = process.stderr
        except FileNotFoundError as ex:
            error = ex

        runnable = False
        version = None
        up_to_date = False
        # Is mpm runnable as-is with provided CLI arguments?
        if not process.returncode and not error:
            runnable = True
            # This regular expression is designed to extract the version number,
            # whether it is surrounded by ANSI color escape sequence or not.
            match = re.compile(
                r"""
                .+                      # Any string
                \                       # A space
                version                 # The "version" string
                \                       # A space
                [^\.]*?                 # Any minimal (non-greedy) string without a dot
                (?P<version>[0-9]+(?:\.[0-9]+)+)   # Version composed of numbers and dots
                .*?                     # Any trailing string (ANSI codes, .dev suffix, etc.)
                $                       # End of the string
                """,
                re.VERBOSE | re.MULTILINE,
            ).search(process.stdout)
            if match:
                version = self.str_to_version(match.groupdict()["version"])
                # Is mpm too old?
                if version >= MPM_MIN_VERSION:
                    up_to_date = True

        return runnable, up_to_date, version, error

    @cached_property
    def ranked_mpm(
        self,
    ) -> list[
        tuple[
            tuple[str, ...], bool, bool, tuple[int, ...] | None, str | Exception | None
        ]
    ]:
        """Rank the mpm candidates we found on the system.

        Sort them by:
        - runnability
        - up-to-date status
        - version number
        - error

        On tie, the order from ``search_mpm`` is respected.
        """
        all_mpm = (
            (mpm_candidate, self.check_mpm(mpm_candidate))
            for mpm_candidate in self.search_mpm()
        )
        return [
            (mpm_args, *mpm_status)
            for mpm_args, mpm_status in sorted(all_mpm, key=itemgetter(1), reverse=True)
        ]

    @cached_property
    def best_mpm(
        self,
    ) -> tuple[
        tuple[str, ...], bool, bool, tuple[int, ...] | None, str | Exception | None
    ]:
        return self.ranked_mpm[0]

    @staticmethod
    def pp(label: str, *args: str) -> None:
        """Print one menu-line with the Xbar/SwiftBar dialect.

        First argument is the menu-line label, separated by a pipe to all other non-
        empty parameters, themselves separated by a space.

        Skip printing of the line if label is empty.
        """
        if label.strip():
            print(
                # Do not strip the label to keep character alignments, especially in
                # table rendering and Python tracebacks.
                label,
                "|",
                *(line for line in map(methodcaller("strip"), args) if line),
                sep=" ",
            )

    @staticmethod
    def print_error_header() -> None:
        """Generic header for blocking error."""
        MPMPlugin.pp("❗️", "dropdown=false")
        print("---")

    def print_error(self, message: str | Exception, submenu: str = "") -> None:
        """Print a formatted error message line by line.

        A red, fixed-width font is used to preserve traceback and exception layout. For
        compactness, the block message is dedented and empty lines are skipped.

        Message is always casted to a string as we allow passing of exception objects
        and have them rendered.
        """
        for line in map(methodcaller("rstrip"), dedent(str(message)).splitlines()):
            if line:
                self.pp(
                    f"{submenu}{line}",
                    self.error_font,
                    "trim=false",
                    "ansi=false",
                    "emojize=false",
                    "symbolize=false" if self.is_swiftbar else "",
                )

    def print_menu(self) -> None:
        """Print the main menu."""
        # Check if we have a recent version of SwiftBar.
        # XXX Xbar does not provide yet a version number in the environment variables.
        if self.is_swiftbar:
            swiftbar_version_str = self.getenv_str("SWIFTBAR_VERSION", "")
            swiftbar_version = self.str_to_version(swiftbar_version_str)
            if not swiftbar_version or swiftbar_version < SWIFTBAR_MIN_VERSION:
                self.print_error_header()
                self.print_error(
                    f"SwiftBar v{swiftbar_version_str} found, but "
                    f"v{self.version_to_str(SWIFTBAR_MIN_VERSION)} is required.",
                )
                return

        # Check if we have a recent version of mpm.
        mpm_args, runnable, up_to_date, _version, error = self.best_mpm
        if not runnable or not up_to_date:
            self.print_error_header()
            if error:
                self.print_error(error)
                print("---")
            action_msg = "Install" if not runnable else "Upgrade"
            min_version_str = self.version_to_str(MPM_MIN_VERSION)
            self.pp(
                f"{action_msg} mpm >= v{min_version_str}",
                f"shell={self.all_pythons[0]}",
                "param1=-m",
                "param2=pip",
                "param3=install",
                "param4=--upgrade",
                # XXX This seems broken beyond repair. No amount of workaround works.
                # See:
                # https://github.com/matryer/xbar/issues/831
                # https://github.com/swiftbar/SwiftBar/issues/308
                # Fallback to the only version that is working on SwiftBar.
                f'param5=\\"meta-package-manager>={min_version_str}\\"',
                self.error_font,
                "refresh=true",
                "terminal=true",
            )
            return

        # Force a sync of all local package databases.
        run((*mpm_args, "--verbosity", "ERROR", "sync"), check=False)

        # Fetch outdated packages from all package managers available on the system.
        # We defer all rendering to mpm itself so it can compute more intricate layouts.
        process = run(
            # We silence all errors but the CRITICAL ones. All others will be captured
            # by mpm in --plugin-output mode and rendered back into each manager
            # section.
            (*mpm_args, "--verbosity", "CRITICAL", "outdated", "--plugin-output"),
            capture_output=True,
            encoding="utf-8",
            check=False,
        )

        # Bail-out immediately on errors related to mpm self-execution or if mpm is
        # not able to produce any output.
        if process.stderr or not process.stdout:
            self.print_error_header()
            self.print_error(process.stderr)
            return

        # Capturing the output of mpm and re-printing it will introduce an extra
        # line returns, hence the extra rstrip() call.
        print(process.stdout.rstrip())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--search-mpm",
        action="store_true",
        help="Locate all mpm on the system and sort them by best candidates.",
    )
    args = parser.parse_args()

    plugin = MPMPlugin()

    if args.search_mpm:
        for mpm_args, runnable, up_to_date, version, error in plugin.ranked_mpm:
            print(
                f"{' '.join(mpm_args)} | runnable: {runnable} | "
                f"up to date: {up_to_date} | version: {version} | error: {error!r}"
            )

    else:
        plugin.print_menu()



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import re

from extra_platforms import UNIX_WITHOUT_MACOS

from ..base import PackageManager
from ..capabilities import search_capabilities, version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class OPKG(PackageManager):
    homepage_url = "https://git.yoctoproject.org/cgit/cgit.cgi/opkg/"

    platforms = UNIX_WITHOUT_MACOS

    requirement = "0.2.0"

    _INSTALLED_REGEXP = re.compile(r"(\S+) - (\S+)")
    _OUTDATED_REGEXP = re.compile(r"(\S+) - (\S+) - (\S+)")
    _SEARCH_REGEXP = re.compile(
        r"""
        (?P<package_id>\S+)
        \ -\
        (?P<version>\S+)
        \ -\
        (?P<description>.+)
        """,
        re.VERBOSE | re.MULTILINE,
    )

    version_regexes = (r"opkg\s+version\s+(?P<version>\S+)",)
    """
    .. code-block:: shell-session

        $ opkg --version
        opkg version 0.3.6 (libsolv 0.7.5)
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ opkg list-installed
            3rd-party-feed-configs - 1.1-r0
            aio-grab - 1.0+git71+c79e264-r0
            alsa-conf - 1.1.9-r0
            alsa-state - 0.2.0-r5
            alsa-states - 0.2.0-r5
            alsa-utils-alsactl - 1.1.9-r0
            avahi-daemon - 0.7-r0
            base-files - 3.0.14-r89
            base-files-dev - 3.0.14-r89
            base-passwd - 3.5.29-r0
            bash - 5.0-r0
            bash-completion - 2.9-r0
            bash-completion-dev - 2.9-r0
            bash-dev - 5.0-r0
            binutils - 2.32.0-r0
            busybox - 1.31.0-r0
            busybox-inetd - 1.31.0-r0
            busybox-mdev - 1.31.0-r0
            busybox-syslog - 1.31.0-r0
            busybox-udhcpc - 1.31.0-r0
        """
        output = self.run_cli("list-installed")

        for package in output.splitlines():
            match = self._INSTALLED_REGEXP.match(package)
            if match:
                package_id, installed_version = match.groups()
                yield self.package(id=package_id, installed_version=installed_version)

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ opkg list-upgradable
            openpli-bootlogo - 20190717-r0 - 20190718-r0
            enigma2-hotplug - 2.7+git1720+55c6b34-r0 - 2.7+git1722+daf2f52-r0
        """
        output = self.run_cli("list-upgradable")

        for package in output.splitlines():
            match = self._OUTDATED_REGEXP.match(package)
            if match:
                package_id, installed_version, latest_version = match.groups()
                yield self.package(
                    id=package_id,
                    latest_version=latest_version,
                    installed_version=installed_version,
                )

    @search_capabilities(extended_support=False, exact_support=False)
    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        .. warning::
            There is no search command so we simulate it by listing all packages.

        .. caution::
            Search does not support extended or exact matching. So we returns the best
            subset of results and let
            :py:meth:`meta_package_manager.base.PackageManager.refiltered_search` refine
            them.

        .. code-block:: shell-session

            $ opkg list
        """
        output = self.run_cli("list")

        for package_id, version, description in self._SEARCH_REGEXP.findall(output):
            yield self.package(
                id=package_id,
                description=description,
                latest_version=version,
            )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ opkg install enigma2-hotplug
        """
        return self.run_cli("install", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ opkg upgrade
        """
        return self.build_cli("upgrade")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ opkg upgrade enigma2-hotplug
        """
        return self.build_cli("upgrade", package_id)

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            $ opkg update
        """
        self.run_cli("update")



# Copyright Kevin Deldycke <kevin@deldycke.com> and contributors.
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from __future__ import annotations

import json
import re

from extra_platforms import UNIX

from ..base import PackageManager
from ..capabilities import version_not_implemented

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Iterator

    from ..base import Package


class PKG(PackageManager):
    name = "FreeBSD System Manager"

    homepage_url = "https://github.com/freebsd/pkg"

    platforms = UNIX

    requirement = "1.11"
    """1.11 is the first version to support ``IGNORE_OSVERSION`` environment variable."""

    pre_args = ("--quiet",)

    _INSTALLED_REGEXP = re.compile(r"(\S+) (\S+) (.+)")
    _OUTDATED_REGEXP = re.compile(r"(\S+): (\S+) -> (\S+) .+")

    """
    .. code-block:: shell-session
        $ pkg --version
        1.20.9
    """

    @property
    def installed(self) -> Iterator[Package]:
        """Fetch installed packages.

        .. code-block:: shell-session

            $ pkg query -e "%a = 0" "%n %v %c"
            7-zip 21.07_2 Console version of the 7-Zip file archiver
            ap24-mod_mpm_itk 2.4.7_2 Run each vhost under a separate uid and gid
            apache24 2.4.57 Version 2.4.x of Apache web server
            aquantia-atlantic-kmod 0.0.5_1 Aquantia AQtion (Atlantic) Network Driver
            arcconf 3.07.23971,1 Adaptec SCSI/SAS RAID administration tool
            areca-cli-amd64 1.14.7.150519,1 Command Line Interface for ARC-xxxx RAID
            base64 1.5_1 Utility to encode and decode base64 files
            bash 5.1.12 GNU Project's Bourne Again SHell
            beadm 1.4_1 Solaris-like utility to manage Boot Environments on ZFS
        """
        output = self.run_cli("query", "-e", r'"%a = 0"', r'"%n %v %c"')

        for package in output.splitlines():
            match = self._INSTALLED_REGEXP.match(package)
            if match:
                package_id, installed_version, description = match.groups()
                yield self.package(
                    id=package_id,
                    description=description,
                    installed_version=installed_version,
                )

    @property
    def outdated(self) -> Iterator[Package]:
        """Fetch outdated packages.

        .. code-block:: shell-session

            $ pkg upgrade --dry-run
            Updating FreeBSD repository catalogue...
            FreeBSD repository is up to date.
            All repositories are up to date.
            Checking for upgrades (312 candidates): 100%
            Processing candidates (312 candidates): 100%
            The following 466 package(s) will be affected (of 0 checked):

            Installed packages to be REMOVED:
                freenas-files: 13.0_1700495253
                py39-midcli: 20190509171453
                py39-middlewared: 13.0_1700495253

            New packages to be INSTALLED:
                abseil: 20230125.3 [FreeBSD]
                argp-standalone: 1.5.0 [FreeBSD]
                brotli: 1.1.0,1 [FreeBSD]

            Installed packages to be UPGRADED:
                7-zip: 21.07_2 -> 23.01 [FreeBSD]
                apache24: 2.4.57 -> 2.4.58_1 [FreeBSD]
                apr: 1.7.0.1.6.1_1 -> 1.7.3.1.6.3_1 [FreeBSD]
                aquantia-atlantic-kmod: 0.0.5_1 -> 0.0.5_2 [FreeBSD]
                bash: 5.1.12 -> 5.2.21 [FreeBSD]

        .. note::

            We rely on ``pkg upgrade`` instead of ``pkg version`` because the latter
            does not provides the new version:

            .. code-block:: shell-session

                $ pkg version --like "<"
                Updating FreeBSD repository catalogue...
                FreeBSD repository is up to date.
                All repositories are up to date.
                7-zip-21.07_2                      <
                apache24-2.4.57                    <
                apr-1.7.0.1.6.1_1                  <
                aquantia-atlantic-kmod-0.0.5_1     <
                bash-5.1.12                        <
        """
        output = self.run_cli("upgrade", "--dry-run")

        outdated_list = output.split("Installed packages to be UPGRADED:", 1)[1].strip()

        for package in outdated_list.splitlines():
            match = self._OUTDATED_REGEXP.match(package.strip())
            if match:
                package_id, installed_version, latest_version = match.groups()
                yield self.package(
                    id=package_id,
                    latest_version=latest_version,
                    installed_version=installed_version,
                )

    def search(self, query: str, extended: bool, exact: bool) -> Iterator[Package]:
        """Fetch matching packages.

        Default search on ID substring:

        .. code-block:: shell-session

            $ pkg search --raw --raw-format json-compact --search name nginx
            {
                "name": "nginx",
                "version": "1.24.0_14,3",
                "comment": "Robust and small WWW server",
                (...)
            }
            {
                "name": "nginx-devel",
                "version": "1.25.3_9",
                "comment": "Robust and small WWW server",
                (...)
            }
            {
                "name": "nginx-ultimate-bad-bot-blocker",
                "version": "4.2020.03.2005_1",
                "comment": "Nginx bad bot and other things blocker",
                (...)
            }
            {
                "name": "p5-Nginx-ReadBody",
                "version": "0.07_1",
                "comment": "Nginx embedded perl module to read a request",
                (...)
            }
            (...)

        Exact search on ID:

        .. code-block:: shell-session

            $ pkg search --raw --raw-format json-compact --search name --exact nginx
            {
                "name": "nginx",
                "origin": "www/nginx",
                "version": "1.24.0_14,3",
                "comment": "Robust and small WWW server",
                "maintainer": "joneum@FreeBSD.org",
                "www": "https://nginx.com/",
                "abi": "FreeBSD:13:amd64",
                "arch": "freebsd:13:x86:64",
                "prefix": "/usr/local",
                "sum": "c39a7696e6eda7bfedba251e4480e50d4c65c520d5a783a584b19b3ef883",
                "flatsize": 1464332,
                "path": "All/nginx-1.24.0_14,3.pkg",
                "repopath": "All/nginx-1.24.0_14,3.pkg",
                "licenselogic": "single",
                "licenses": [
                    "BSD2CLAUSE"
                ],
                "pkgsize": 473632,
                "desc": "NGINX is a high performance edge web server with the (...)",
                "deps": {
                    "pcre2": {
                        "origin": "devel/pcre2",
                        "version": "10.42"
                    }
                },
                "categories": [
                    "www"
                ],
                "shlibs_required": [
                    "libpcre2-8.so.0"
                ],
                "options": {
                    "AJP": "off",
                    "ARRAYVAR": "off",
                    "AWS_AUTH": "off",
                    "BROTLI": "off",
                    "CACHE_PURGE": "off",
                    "CLOJURE": "off",
                    "COOKIE_FLAG": "off",
                    "CT": "off",
                    "DEBUG": "off",
                    "DEBUGLOG": "off",
                    "DEVEL_KIT": "off",
                    "DRIZZLE": "off",
                    "DSO": "on",
                    "DYNAMIC_UPSTREAM": "off",
                    "ECHO": "off",
                    "ENCRYPTSESSION": "off",
                    "FILE_AIO": "on",
                    "FIPS_CHECK": "off",
                    "FORMINPUT": "off",
                    "GOOGLE_PERFTOOLS": "off",
                    "GRIDFS": "off",
                    "GSSAPI_HEIMDAL": "off",
                    "GSSAPI_MIT": "off",
                    "HEADERS_MORE": "off",
                    "HTTP": "on",
                    "HTTPV2": "on",
                    "HTTPV3": "off",
                    "HTTPV3_BORING": "off",
                    "HTTPV3_LSSL": "off",
                    "HTTPV3_QTLS": "off",
                    "HTTP_ACCEPT_LANGUAGE": "off",
                    "HTTP_ADDITION": "on",
                    "HTTP_AUTH_DIGEST": "off",
                    "HTTP_AUTH_KRB5": "off",
                    "HTTP_AUTH_LDAP": "off",
                    "HTTP_AUTH_PAM": "off",
                    "HTTP_AUTH_REQ": "on",
                    "HTTP_CACHE": "on",
                    "HTTP_DAV": "on",
                    "HTTP_DAV_EXT": "off",
                    "HTTP_DEGRADATION": "off",
                    "HTTP_EVAL": "off",
                    "HTTP_FANCYINDEX": "off",
                    "HTTP_SUBS_FILTER": "off",
                    "HTTP_TARANTOOL": "off",
                    "HTTP_UPLOAD": "off",
                    "HTTP_UPLOAD_PROGRESS": "off",
                    "HTTP_UPSTREAM_CHECK": "off",
                    "HTTP_UPSTREAM_FAIR": "off",
                    "HTTP_UPSTREAM_STICKY": "off",
                    "HTTP_VIDEO_THUMBEXTRACTOR": "off",
                    "HTTP_XSLT": "off",
                    "HTTP_ZIP": "off",
                    "ICONV": "off",
                    "IPV6": "on",
                    "LET": "off",
                    "LINK": "off",
                    "LUA": "off",
                    "MAIL": "on",
                    "MAIL_IMAP": "off",
                    "MAIL_POP3": "off",
                    "MAIL_SMTP": "off",
                    "MAIL_SSL": "on",
                    "MEMC": "off",
                    "MODSECURITY3": "off",
                    "NAXSI": "off",
                    "NJS": "off",
                    "NJS_XML": "off",
                    "OPENTRACING": "off",
                    "PASSENGER": "off",
                    "POSTGRES": "off",
                    "RDS_CSV": "off",
                    "RDS_JSON": "off",
                    "REDIS2": "off",
                    "RTMP": "off",
                    "SET_MISC": "off",
                    "SFLOW": "off",
                    "SHIBBOLETH": "off",
                    "SLOWFS_CACHE": "off",
                    "SRCACHE": "off",
                    "STREAM": "on",
                    "STREAM_REALIP": "on",
                    "STREAM_SSL": "on",
                    "STREAM_SSL_PREREAD": "on",
                    "STS": "off",
                    "THREADS": "on",
                    "VOD": "off",
                    "VTS": "off",
                    "WEBSOCKIFY": "off",
                    "WWW": "on",
                    "XSS": "off"
                },
                "annotations": {
                    "FreeBSD_version": "1302001",
                    "build_timestamp": "2024-01-07T10:41:34+0000",
                    "built_by": "poudriere-git-3.4.0",
                    "cpe": "cpe:2.3:a:f5:nginx:1.24.0:::::freebsd13:x64:14",
                    "port_checkout_unclean": "no",
                    "port_git_hash": "756e18783",
                    "ports_top_checkout_unclean": "no",
                    "ports_top_git_hash": "756e18783"
                }
            }

        Extended search:

        .. code-block:: shell-session

            $ pkg search --raw --raw-format json-compact \
              --search name --search comment --search description nginx
        """
        search_args = ["--raw", "--raw-format", "json-compact", "--search", "name"]
        if exact:
            search_args.append("--exact")
        # Expand search to the comment and description fields.
        if extended:
            search_args += ["--search", "comment", "--search", "description"]

        output = self.run_cli(search_args, query)

        for package in map(json.loads, output.splitlines()):
            yield self.package(
                id=package["name"],
                description=package["comment"],
                latest_version=package["version"],
            )

    @version_not_implemented
    def install(self, package_id: str, version: str | None = None) -> str:
        """Install one package.

        .. code-block:: shell-session

            $ pkg install --yes dmg2img
            Updating FreeBSD repository catalogue...
            FreeBSD repository is up to date.
            All repositories are up to date.
            Checking integrity... done (0 conflicting)
            The following 1 package(s) will be affected (of 0 checked):

            New packages to be INSTALLED:
                dmg2img: 1.6.7 [FreeBSD]

            Number of packages to be installed: 1
            [1/1] Installing dmg2img-1.6.7...
            [1/1] Extracting dmg2img-1.6.7: 100%
        """
        return self.run_cli("install", "--yes", package_id)

    def upgrade_all_cli(self) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ pkg upgrade --yes
        """
        return self.build_cli("upgrade", "--yes")

    @version_not_implemented
    def upgrade_one_cli(
        self,
        package_id: str,
        version: str | None = None,
    ) -> tuple[str, ...]:
        """Generates the CLI to upgrade all packages (default) or only the one provided
        as parameter.

        .. code-block:: shell-session

            $ pkg upgrade --yes dmg2img
        """
        return self.build_cli("upgrade", "--yes", package_id)

    def remove(self, package_id: str) -> str:
        """Remove one package.

        .. code-block:: shell-session

            $ pkg delete --yes dmg2img
            Checking integrity... done (0 conflicting)
            Deinstallation has been requested for the following 1 packages:

            Installed packages to be REMOVED:
                dmg2img: 1.6.7

            Number of packages to be removed: 1
            [1/1] Deinstalling dmg2img-1.6.7...
            [1/1] Deleting files for dmg2img-1.6.7: 100%
            pkg: Package database is busy while closing!
        """
        return self.run_cli("delete", "--yes", package_id)

    def sync(self) -> None:
        """Sync package metadata.

        .. code-block:: shell-session

            $ IGNORE_OSVERSION=yes pkg update
            Updating FreeBSD repository catalogue...
            Fetching meta.conf: 100%    163 B   0.2kB/s    00:01
            Fetching packagesite.pkg: 100%    7 MiB   3.6MB/s    00:02
            Processing entries: 100%
            FreeBSD repository update completed. 33804 packages processed.
            All repositories are up to date.

        The ``IGNORE_OSVERSION=yes`` prevents blocking update:

        .. code-block:: shell-session

            $ pkg update
            Updating FreeBSD repository catalogue...
            Fetching meta.conf: 100%    163 B   0.2kB/s    00:01
            Fetching packagesite.pkg: 100%    7 MiB   3.6MB/s    00:02
            Processing entries:   0%
            Newer FreeBSD version for package zziplib:
            To ignore this error set IGNORE_OSVERSION=yes
            - package: 1302001
            - running kernel: 1301000
            Ignore the mismatch and continue? [y/N]:
        """
        self.run_cli("update", override_extra_env={"IGNORE_OSVERSION": "yes"})

    def cleanup(self) -> None:
        """Removes things we don't need anymore.

        .. code-block:: shell-session

            $ pkg autoremove --yes
            Checking integrity... done (0 conflicting)
            Nothing to do.

        .. code-block:: shell-session

            $ pkg clean --yes --all
            Nothing to do.
        """
        self.run_cli("autoremove", "--yes")
        self.run_cli("clean", "--yes", "--all")
